(* 
  Code generation for Gwiz: 
  translate takes a semantically checked AST and produces LLVM IR
  Based on MicroC 
  File: codegen.ml 
  Authors: Ashley Kim, Elisa Luo, Katherine Duff
  PLT Spring 2021

LLVM tutorial: Make sure to read the OCaml version of the tutorial

http://llvm.org/docs/tutorial/index.html

Detailed documentation on the OCaml LLVM library:

http://llvm.moe/
http://llvm.moe/ocaml/

*)



module L = Llvm
module A = Ast
open Sast 

module StringMap = Map.Make(String)

(* translate : Sast.program -> Llvm.module *)
let translate (globals, functions) =
  let context    = L.global_context () in
  
  (* Create the LLVM compilation module into which
     we will generate code *)
  let the_module = L.create_module context "gWiz" in

  (* Get types from the context *)
  let i32_t      = L.i32_type     context
  and i8_t       = L.i8_type      context
  and double_t   = L.double_type  context
  and string_t   = L.pointer_type (L.i8_type context)
  and i1_t       = L.i1_type      context
  and void_t     = L.void_type    context in
  (*Pointers for the array*)
  let i32_ptr_t  = L.pointer_type i32_t 
  and i64_t     = L.i64_type      context in
  let node_t     = let node_t = L.named_struct_type context "Node" in
                    L.struct_set_body node_t [| i64_t ; L.pointer_type node_t |] false;
                    node_t in
  let node_pt = L.pointer_type node_t in
  let arr_t = let arr_t = L.named_struct_type context "Array" in
                    L.struct_set_body arr_t [| node_pt ; i32_t |] false;
                    arr_t in
  let arr_pt = L.pointer_type arr_t         
  in

  let pointer_t = L.pointer_type in 
  let trackpoint_t = L.named_struct_type context "trackpoint_t" in 
  L.struct_set_body trackpoint_t [| double_t; double_t; i32_t|] false; 
  let trk_pointer_t = pointer_t trackpoint_t in 

  (* Return the LLVM type for a gWiz type *)
  let ltype_of_typ = function
      A.Int   -> i32_t
    | A.Double -> double_t
    | A.Void  -> void_t
    | A.String -> string_t
    | A.Bool -> i1_t 
    | A.Trackpoint -> trk_pointer_t
    | A.ArrayType(_) -> arr_pt
    | A.File -> i32_ptr_t
  in

	(*Create a mpa of global variables after creating each *)
	let global_vars: L.llvalue StringMap.t = 
  let global_var m (t,n) = 
    let init = match t with
      (*add other types to pattern match here*)
        A.Double -> L.const_float (ltype_of_typ t) 0.0
      | A.Trackpoint -> L.const_pointer_null (ltype_of_typ t)
      | _ -> L.const_int (ltype_of_typ t) 0
    in StringMap.add n (L.define_global n init the_module) m in
  List.fold_left global_var StringMap.empty globals in

  (* Declare built in print functions *)
  let printf_t : L.lltype = 
      L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
  let printf_func : L.llvalue = 
      L.declare_function "printf" printf_t the_module in

  (* Declare built in array functions *)
  let initlist_t : L.lltype = 
    L.var_arg_function_type arr_pt [| i64_t |] in
  let initlist_func : L.llvalue = 
    L.declare_function "initList" initlist_t the_module in
    
  let addend_t : L.lltype =
    L.var_arg_function_type node_pt [| arr_pt; i64_t |] in
  let addend_func : L.llvalue = 
    L.declare_function "addEnd" addend_t the_module in

  let getelemindex_t : L.lltype = 
    L.var_arg_function_type i64_t [| arr_pt; i32_t |] in
  let getelemindex_func : L.llvalue = 
    L.declare_function "getElemIndex" getelemindex_t the_module in


  (*Declare GPX Parsing functions *)
  let numpt_t: L.lltype =
      L.function_type i32_t [| string_t |] in
  let numpt_func: L.llvalue =
      L.declare_function "numPt" numpt_t the_module in

  let parseGPX_t: L.lltype =
      L.function_type i32_ptr_t [| string_t |] in
  let parseGPX_func: L.llvalue =
      L.declare_function "parseGPX" parseGPX_t the_module in

  let printFirst_t: L.lltype =
      L.function_type i32_t [| i32_ptr_t |] in
  let printFirst_func: L.llvalue =
      L.declare_function "printFirst" printFirst_t the_module in
  
  (*declare built in gpx functions*)
  let totTime_t: L.lltype =
      L.function_type i32_t [| i32_ptr_t |] in
  let totTime_func: L.llvalue =
      L.declare_function "totTime" totTime_t the_module in

  let totDist_t: L.lltype =
      L.function_type double_t [| i32_ptr_t |] in
  let totDist_func: L.llvalue =
      L.declare_function "totDist" totDist_t the_module in

  let avgSpeed_t: L.lltype =
      L.function_type double_t [| i32_ptr_t |] in
  let avgSpeed_func: L.llvalue =
      L.declare_function "avgSpeed" avgSpeed_t the_module in

  let stat_t: L.lltype =
      L.function_type i32_t [| i32_ptr_t |] in
  let stat_func: L.llvalue =
      L.declare_function "stat" stat_t the_module in


  (*declare built in math functions*)
  let intFloor_t: L.lltype =
      L.function_type i32_t [| double_t |] in
  let intFloor_func: L.llvalue =
      L.declare_function "intFloor" intFloor_t the_module in

  let squareRoot_t: L.lltype =
      L.function_type double_t [| double_t |] in
  let squareRoot_func: L.llvalue =
      L.declare_function "squareRoot" squareRoot_t the_module in

  let doubleFloor_t: L.lltype =
      L.function_type double_t [| double_t |] in
  let doubleFloor_func: L.llvalue =
      L.declare_function "doubleFloor" doubleFloor_t the_module in

  (* Linked getter functions for Trackpoint *)
  let getLat_t : L.lltype = 
    L.function_type double_t [|trk_pointer_t|] in 
  let getLat_func : L.llvalue = 
    L.declare_function "getLat" getLat_t the_module in 

  let getLon_t : L.lltype = 
    L.function_type double_t [|trk_pointer_t|] in 
  let getLon_func : L.llvalue = 
    L.declare_function "getLon" getLon_t the_module in 

  let getTime_t : L.lltype = 
    L.function_type i32_t [|trk_pointer_t|] in 
  let getTime_func : L.llvalue = 
    L.declare_function "getTime" getTime_t the_module in 

  (* Ensures int helper *)
  let ensureInt c = 
    if L.type_of c = double_t then (L.const_fptosi c i32_t) else c in
    
  (* Ensures double helper *)
  let ensureDouble c = 
    if L.type_of c = double_t then c else (L.const_sitofp c double_t) in

  (* Define each function (arguments and return type) so we can 
     call it even before we've created its body *)
  let function_decls : (L.llvalue * sfunc_decl) StringMap.t =
    let function_decl m fdecl =
      let name = fdecl.sfname
      and formal_types = 
	Array.of_list (List.map (fun (t,_) -> ltype_of_typ t) fdecl.sformals)
      in let ftype = L.function_type (ltype_of_typ fdecl.styp) formal_types in
      StringMap.add name (L.define_function name ftype the_module, fdecl) m in
    List.fold_left function_decl StringMap.empty functions in
  
  (* Fill in the body of the given function *)
  let build_function_body fdecl =
    let (the_function, _) = StringMap.find fdecl.sfname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in

    let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder
    and string_format_str = L.build_global_stringptr "%s\n" "fmt" builder 
    and double_format_str = L.build_global_stringptr "%g\n" "fmt" builder 
    and bool_format_str   = L.build_global_stringptr "%d\n" "fmt" builder
   in

	(*Construct the function's "locals": formal arguments and locally declared variables. 
	Allocate each on the stack, initialize their value, if appropriate, and remember their 
	values in the "locals" map *)
	let local_vars =
		let add_formal m (t, n) p = 
			let local = L.build_alloca (ltype_of_typ t) n builder in 
			ignore (L.build_store p local builder);
			StringMap.add n local m
		and add_local m (t, n) = 
			let local_var = L.build_alloca (ltype_of_typ t) n builder
			in StringMap.add n local_var m in
		
		let formals = List.fold_left2 add_formal StringMap.empty
		  fdecl.sformals (Array.to_list (L.params the_function)) in 
		List.fold_left add_local formals fdecl.slocals in
  
	(*Return the value for a variable or formal argument. Check local names first, then
	global names*)
	let lookup n = try StringMap.find n local_vars 
		with Not_found -> StringMap.find n global_vars in

    (* Construct code for an expression; return its value *)
    let rec expr builder ((t, e) : sexpr) = match e with
	      SLiteral i  -> L.const_int i32_t i
      | SDLit l     -> L.const_float_of_string double_t l
      | SBLit b     -> L.const_int i1_t (if b then 1 else 0) 
      | SSLit s     -> L.build_global_stringptr s "str" builder 
      | SNoexpr     -> L.const_int i32_t 0
      | SId s       -> L.build_load (lookup s) s builder
      | STrackpointLit (e1, e2, e3) ->
        let ttmp = L.build_alloca trackpoint_t "trackpoint_tmp" builder in 
        let tptr = L.build_alloca (L.pointer_type trackpoint_t) "trackpoint_pointer" builder in 
        
        let e1' = ensureDouble (expr builder e1)
        and e2' = ensureDouble (expr builder e2)
        and e3' = ensureInt (expr builder e3) in 

        let lat_tmp = L.build_struct_gep ttmp 0 "lat" builder in 
          ignore (L.build_store e1' lat_tmp builder); 
        let lon_tmp = L.build_struct_gep ttmp 1 "lon" builder in 
          ignore (L.build_store e2' lon_tmp builder);
        let time_tmp = L.build_struct_gep ttmp 2 "time" builder in 
          ignore (L.build_store e3' time_tmp builder);
          ignore (L.build_store ttmp tptr builder); 
        let trkld = L.build_load tptr "" builder in trkld 
      | STrackpointAccess (e1, s2) -> 
        let trackpoint = expr builder e1 in
        (match s2 with 
        | "lat" -> L.build_call getLat_func [|trackpoint|] "getLat" builder
        | "lon" -> L.build_call getLon_func [|trackpoint|] "getLon" builder 
        | "time" -> L.build_call getTime_func [|trackpoint|] "getTime" builder 
        | _ -> raise (Failure ("Attribute does not exist"))
        )
      | SArrayLiteral l -> build_arr builder t l 
      | SArrayAccess (array, index) ->
          let (t, _) = array in
            let check_arr_type = match t with
            A.ArrayType t -> t
            | _ -> raise(Failure("Can only access an array")) in 
          let var = L.build_call getelemindex_func [| expr builder array; expr builder index|] "getElement" builder in
          (match check_arr_type with
              A.Int -> L.build_trunc var i32_t "arrayAcc" builder
              | A.Trackpoint -> L.build_inttoptr var trk_pointer_t "arrayAcc" builder
              | _ -> raise(Failure("Accessing illegal array type"))
          )
      | SArrayAppend (array, element) ->
          let(t1, _) = array in
          let check_arr_type =
            match t1 with
              A.ArrayType t -> t 
              | _ -> raise(Failure("can only append to an array type")) in
          let (elem_typ, _) = element in
          if check_arr_type <> elem_typ then raise(Failure("element type must be the same as array type"))
          else 
          let built_expr = expr builder element in
          let t = L.type_of built_expr in
          let data = match L.classify_type t with
            L.TypeKind.Pointer -> L.build_pointercast built_expr i64_t "" builder
            | L.TypeKind.Integer -> L.build_zext built_expr i64_t "" builder
            | _ -> raise (Failure("cannot convert array type "))
        in L.build_call addend_func [| expr builder array; data |] "addEnd" builder
    
      (*| STrackArrayLiteral(l) ->
            (*initialize array*)
            let list_size = List.length l in
            let ll_listsize =  L.const_int i32_t list_size in
            let list_hd = (List.nth l 0) in
            let (ast_typ, other) = list_hd in
            let match_typ = match ast_typ with
               A.Trackpoint -> A.Trackpoint
              | _ -> raise(Failure("Track list must be Trackpoint")) in
            let init_array = L.build_array_malloc (trk_pointer_t) ll_listsize "" builder in
            
            (* add the list elements *)
            let elements = List.map (fun e -> expr builder e) l in 
            ignore(List.fold_left (fun i elem ->
              let idx = L.const_int i32_t i in
              let elemptr = L.build_gep init_array [| idx |] "" builder in
              let a = L.build_store elem elemptr builder in
             i+1) 0 elements); init_array

      | STrackArrayAccess (array, index) ->
          let arr_var = expr builder array in
          let idx = expr builder index in 

          let ptr = L.build_gep arr_var [| idx |] "" builder in
          let cast_pt = L.build_ptrtoint ptr i64_t "" builder in
          let tpt = L.build_pointercast cast_pt trk_pointer_t "" builder in

          L.build_load tpt "" builder 

      | STrackArrayInit (tarrtype, siz) ->
          let len = expr builder siz in 
         let init_array = L.build_array_malloc (trk_pointer_t) len "" builder in 
         (* let init_array = L.build_array_alloc (trackpoint_t) len "" builder in *)
          init_array *)

          
      | SAssign (s, e) -> let e' = expr builder e in 
                          ignore(L.build_store e' (lookup s) builder); e'
      | SBinop ((A.Double,_ ) as e1, op, e2) ->
          let e1' = expr builder e1
          and e2' = expr builder e2 in
          (match op with 
            A.Add     -> L.build_fadd
          | A.Sub     -> L.build_fsub
          | A.Mult    -> L.build_fmul
          | A.Div     -> L.build_fdiv 
          | A.Equal   -> L.build_fcmp L.Fcmp.Oeq
          | A.Neq     -> L.build_fcmp L.Fcmp.One
          | A.Less    -> L.build_fcmp L.Fcmp.Olt
          | A.Leq     -> L.build_fcmp L.Fcmp.Ole
          | A.Greater -> L.build_fcmp L.Fcmp.Ogt
          | A.Geq     -> L.build_fcmp L.Fcmp.Oge
          | A.And | A.Or ->
              raise (Failure "internal error: semant should have rejected and/or on double")
          ) e1' e2' "tmp" builder
            | SBinop (e1, op, e2) ->
          let e1' = expr builder e1
          and e2' = expr builder e2 in
          (match op with
            A.Add     -> L.build_add
          | A.Sub     -> L.build_sub
          | A.Mult    -> L.build_mul
          | A.Div     -> L.build_sdiv
          | A.And     -> L.build_and
          | A.Or      -> L.build_or
          | A.Equal   -> L.build_icmp L.Icmp.Eq
          | A.Neq     -> L.build_icmp L.Icmp.Ne
          | A.Less    -> L.build_icmp L.Icmp.Slt
          | A.Leq     -> L.build_icmp L.Icmp.Sle
          | A.Greater -> L.build_icmp L.Icmp.Sgt
          | A.Geq     -> L.build_icmp L.Icmp.Sge
          ) e1' e2' "tmp" builder
      | SUnop(op, ((t, _) as e)) ->
          let e' = expr builder e in
	        (match op with
	          A.Neg when t = A.Double -> L.build_fneg 
	        | A.Neg                  -> L.build_neg
          | A.Not                  -> L.build_not) e' "tmp" builder
      | SCall ("printi", [e]) ->
	      L.build_call printf_func [| int_format_str ; (expr builder e) |]
	      "printf" builder
      | SCall ("prints", [e]) -> 
        L.build_call printf_func [| string_format_str ; (expr builder e) |]
        "printf" builder
      | SCall ("printd", [e]) -> 
	      L.build_call printf_func [| double_format_str ; (expr builder e) |]
	      "printf" builder
      | SCall ("printb", [e]) -> 
	      L.build_call printf_func [| bool_format_str ; (expr builder e) |]
	      "printf" builder
      | SCall ("numPt", [e]) ->
          L.build_call numpt_func [|(expr builder e)|] "numPt" builder
      | SCall ("printFirst", [e]) ->
          L.build_call printFirst_func [|(expr builder e)|] "printFirst" builder
      | SCall ("stat", [e]) ->
          L.build_call stat_func [|(expr builder e)|] "stat" builder
      |SCall ("totTime", [e]) ->
          L.build_call totTime_func [|(expr builder e)|] "totTime" builder
      |SCall ("totDist", [e]) ->
          L.build_call totDist_func [|(expr builder e)|] "totDist" builder
      |SCall ("avgSpeed", [e]) ->
          L.build_call avgSpeed_func [|(expr builder e)|] "avgSpeed" builder
      |SCall ("intFloor", [e]) ->
          L.build_call intFloor_func [|(expr builder e)|] "intFloor" builder
      |SCall ("squareRoot", [e]) ->
          L.build_call squareRoot_func [|(expr builder e)|] "squareRoot" builder
      |SCall ("doubleFloor", [e]) ->
          L.build_call doubleFloor_func [|(expr builder e)|] "doubleFloor" builder
      |SCall ("parseGPX", [e]) ->
          L.build_call parseGPX_func [|(expr builder e)|] "parseGPX" builder
      | SCall("getLat", [e]) -> 
          L.build_call getLat_func [| (expr builder e)|] "getLat" builder 
      | SCall("getLon", [e]) -> 
          L.build_call getLon_func [| (expr builder e)|] "getLon" builder 
      | SCall("getTime", [e]) -> 
          L.build_call getTime_func [| (expr builder e)|] "getTime" builder
  | SCall (f, args) ->
         let (fdef, fdecl) = StringMap.find f function_decls in
	    let llargs = List.rev (List.map (expr builder) (List.rev args)) in
	    let result = (match fdecl.styp with 
                        A.Void -> ""
                      | _ -> f ^ "_result") in
         L.build_call fdef (Array.of_list llargs) result builder
    

  (* Functions to initialize and build Array Literal*)
  
  and element_type = function
    A.ArrayType t -> t 
    | _ -> raise (Failure("need array type"))

  and element_size t = 
    let typ_size = element_type t in 
    L.size_of(ltype_of_typ typ_size)

  and build_arr builder t arr = 
    let type_size = element_size t in
    let list = L.build_call initlist_func [| type_size |] "initList" builder in

    let add_end e = 
      let built_expr = expr builder e in
      let t = L.type_of built_expr in
      let data = match L.classify_type t with
        L.TypeKind.Pointer -> L.build_pointercast built_expr i64_t "" builder
        | L.TypeKind.Integer -> L.build_zext built_expr i64_t "" builder
        | _ -> raise (Failure("cannot convert array type"))
    in
    ignore(L.build_call addend_func [| list; data |] "addEnd" builder) 
  in List.iter add_end arr; list
in
    
    (* LLVM insists each basic block end with exactly one "terminator" 
       instruction that transfers control.  This function runs "instr builder"
       if the current block does not already have a terminator.  Used,
       e.g., to handle the "fall off the end of the function" case. *)
    let add_terminal builder instr =
      match L.block_terminator (L.insertion_block builder) with
	Some _ -> ()
      | None -> ignore (instr builder) in
	
    (* Build the code for the given statement; return the builder for
       the statement's successor (i.e., the next instruction will be built
       after the one generated by this call) *)

    let rec stmt builder = function
	      SBlock sl -> List.fold_left stmt builder sl
      | SExpr e -> ignore(expr builder e); builder 
      | SReturn e -> ignore(match fdecl.styp with
                              (* Special "return nothing" instr *)
                              A.Void -> L.build_ret_void builder 
                              (* Build return statement *)
                            | _ -> L.build_ret (expr builder e) builder );
                     builder
      | SIf (predicate, then_stmt, else_stmt) ->
        let bool_val = expr builder predicate in
        (* Add merge block to the list of blocks for the function *)
	      let merge_bb = L.append_block context "merge" the_function in
        (* Partial function to create branch with merge block *)
        let build_br_merge = L.build_br merge_bb in 

        (* "then" block *)
        let then_bb = L.append_block context "then" the_function in
        add_terminal (stmt (L.builder_at_end context then_bb) then_stmt)
        build_br_merge;

        (* "else block" *)
        let else_bb = L.append_block context "else" the_function in
        add_terminal (stmt (L.builder_at_end context else_bb) else_stmt)
        build_br_merge;

        ignore(L.build_cond_br bool_val then_bb else_bb builder);
        L.builder_at_end context merge_bb

      | SWhile (predicate, body) ->
        let pred_bb = L.append_block context "while" the_function in
        ignore(L.build_br pred_bb builder);
      
        let body_bb = L.append_block context "while_body" the_function in
        add_terminal (stmt (L.builder_at_end context body_bb) body)
          (L.build_br pred_bb);
    
        let pred_builder = L.builder_at_end context pred_bb in
        let bool_val = expr pred_builder predicate in
      
        let merge_bb = L.append_block context "merge" the_function in
        ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
        L.builder_at_end context merge_bb
      
      (* Implement for loops as while loops *)
      | SFor (e1, e2, e3, body) -> stmt builder
        ( SBlock [SExpr e1 ; SWhile (e2, SBlock [body ; SExpr e3]) ] )

    in

    (* Build the code for each statement in the function *)
    let builder = stmt builder (SBlock fdecl.sbody) in

    (* Add a return if the last block falls off the end *)
    add_terminal builder (match fdecl.styp with
        A.Void -> L.build_ret_void
      | A.Double -> L.build_ret(L.const_float double_t 0.0)
      | A.Int -> L.build_ret (L.const_int i32_t 0)
      | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
  in

  List.iter build_function_body functions;
  the_module
