(* 
  Semantic checking for the GWiz compiler
  Based on MicroC 
  File: semant.ml 
  Authors: Ashley Kim, Katherine Duff
  PLT Spring 2021 
*)

open Ast
open Sast

module StringMap = Map.Make(String)

(* Semantic checking of the AST. Returns an SAST if successful,
   throws an exception if something is wrong.

   Check each global variable, then check each function *)

let check (globals, functions) =
 
  
  (* Verify a list of bindings has no void types or duplicate names *)
  let check_binds (kind : string) (binds : bind list) =
    List.iter (function
	(Void, b) -> raise (Failure ("illegal void " ^ kind ^ " " ^ b))
      | _ -> ()) binds;
    let rec dups = function
        [] -> ()
      |	((_,n1) :: (_,n2) :: _) when n1 = n2 ->
	  raise (Failure ("duplicate " ^ kind ^ " " ^ n1))
      | _ :: t -> dups t
    in dups (List.sort (fun (_,a) (_,b) -> compare a b) binds)
  in

  (**** Check global variables ****)

  check_binds "global" globals;

  (**** Check functions ****)
  
  (* Collect function declarations for built-in functions: no bodies *)
  let built_in_decls = 
    let add_bind map (ftyp, name, ty) = StringMap.add name {
      typ = ftyp;
      fname = name; 
      formals = [(ty, "x")];
      locals = []; body = [] } map
    in List.fold_left add_bind StringMap.empty [ (Void,"printi", Int); 
                               (Void,"prints", String);
                               (Void,"printd", Double); 
                               (Void,"printb", Bool);
                               (Int, "numPt", String);
                               (Int, "totTime", File);
                               (Double, "totDist", File);
                               (Double, "avgSpeed", File);
                               (File, "parseGPX", String);
                               (Void, "printFirst", File);
                               (Void, "stat", File);
                               (Int, "intFloor", Double);
                               (Double, "doubleFloor", Double);
                               (Double, "squareRoot", Double);
                               (Double,"getLat", Trackpoint); 
                               (Double,"getLon", Trackpoint);
                               (Int, "getTime", Trackpoint); 
                                              ]

  in

  (* Add function name to symbol table *)
  let add_func map fd = 
    let built_in_err = "function " ^ fd.fname ^ " may not be defined"
    and dup_err = "duplicate function " ^ fd.fname
    and make_err er = raise (Failure er)
    and n = fd.fname (* Name of the function *)
    in match fd with (* No duplicate functions or redefinitions of built-ins *)
         _ when StringMap.mem n built_in_decls -> make_err built_in_err
       | _ when StringMap.mem n map -> make_err dup_err  
       | _ ->  StringMap.add n fd map 
  in

  (* Collect all function names into one symbol table *)
  let function_decls = List.fold_left add_func built_in_decls functions
  in
  
  (* Return a function from our symbol table *)
  let find_func s = 
    try StringMap.find s function_decls
    with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  (*let _ = find_func "main" in (* Ensure "main" is defined *)*)

  
  let check_function func =
    (* Make sure no formals or locals are void or duplicates *)
    check_binds "formal" func.formals;
    check_binds "local" func.locals;
    

    (* Raise an exception if the given rvalue type cannot be assigned to
       the given lvalue type *)
    let check_assign lvaluet rvaluet err =
       if lvaluet = rvaluet then lvaluet else raise (Failure ("Assign error in semant " ^ err))
    in   

    (* Build local symbol table of variables for this function *)
    let symbols = List.fold_left (fun m (ty, name) -> StringMap.add name ty m)
	                StringMap.empty (globals @ func.formals @ func.locals )
    in

    (* Return a variable from our local symbol table *)
    let type_of_identifier s =
      try StringMap.find s symbols
      with Not_found -> raise (Failure ("undeclared identifier " ^ s))
    in

    (* Return a semantically-checked expression, i.e., with a type *)
    let rec expr = function
    Literal  l -> (Int, SLiteral l)
  | DLit l -> (Double, SDLit l)
  | SLit l -> (String, SSLit l)
  | BLit l  -> (Bool, SBLit l)
  | Noexpr     -> (Void, SNoexpr)
  | Id s       -> (type_of_identifier s, SId s)
  | TrackpointLit(e1, e2, e3) -> (* check lat, long, time are correct types *)
      let (t1, _) = expr e1
      and (t2, _) = expr e2
      and (t3, _) = expr e3 in 

      if t1 = Double && t2 = Double && t3 = Int then (Trackpoint, STrackpointLit (expr e1, expr e2, expr e3))
      else raise (Failure ("expected numeric trackpoint args")) 
  | TrackpointAccess(_,a) -> 
      (match a with 
      | "lat" | "lon" | "time" -> (Double, SDLit a)
      | _ -> raise(Failure ("error semant access type"))
      ) 
  | Assign(var, e) as ex -> 
      let lt = type_of_identifier var
      and (rt, e') = expr e in
      let err = "illegal assignment " ^ string_of_typ lt ^ " = " ^ 
        string_of_typ rt ^ " in " ^ string_of_expr ex
      in (check_assign lt rt err, SAssign(var, (rt, e')))
  | Unop(op, e) as ex -> 
      let (t, e') = expr e in
      let ty = match op with
        Neg when t = Int || t = Double -> t 
      | Not when t = Bool -> Bool
      | _ -> raise (Failure ("illegal unary operator " ^ 
                             string_of_uop op ^ string_of_typ t ^
                             " in " ^ string_of_expr ex))
      in (ty, SUnop(op, (t, e')))
  | Binop(e1, op, e2) as e -> 
      let (t1, e1') = expr e1 
      and (t2, e2') = expr e2 in
      (* All binary operators require operands of the same type *)
      let same = t1 = t2 in
      (* Determine expression type based on operator and operand types *)
      let ty = match op with
        Add | Sub | Mult | Div when same && t1 = Int   -> Int
      | Add | Sub | Mult | Div when same && t1 = Double -> Double 
      | Equal | Neq            when same               -> Bool
      | Less | Leq | Greater | Geq
                 when same && (t1 = Int || t1 = Double) -> Bool
      | And | Or when same && t1 = Bool -> Bool
      | _ -> raise (
    Failure ("illegal binary operator " ^
                   string_of_typ t1 ^ " " ^ string_of_op op ^ " " ^
                   string_of_typ t2 ^ " in " ^ string_of_expr e))
      in (ty, SBinop((t1, e1'), op, (t2, e2')))
  | Call(fname, args) as call -> 
      let fd = find_func fname in
      let param_length = List.length fd.formals in
      if List.length args != param_length then
        raise (Failure ("expecting " ^ string_of_int param_length ^ 
                        " arguments in " ^ string_of_expr call))
      else let check_call (ft, _) e = 
        let (et, e') = expr e in 
        let err = "illegal argument found " ^ string_of_typ et ^
          " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e
        in (check_assign ft et err, e')
      in 
      let args' = List.map2 check_call fd.formals args
      in (fd.typ, SCall(fname, args'))
  | ArrayLiteral(l) -> 
      if List.length l > 0 then 
        let typ = expr(List.nth l 0) in
        let (arraytype, _) = typ in
        let check_array e =
          let (et, e') = expr e in (et, e')
          in let l' = List.map check_array l in
          let types e = 
            let (a, _)  = expr e in 
            if a <> arraytype then 
                    raise(Failure("array of different types" ^ string_of_typ a ^ " is not " ^ string_of_typ arraytype))
            in let _ = List.map types l in (ArrayType(arraytype), SArrayLiteral(l'))

      else raise (Failure ("Must have at least one element in list upon initialization"))
  | ArrayAccess (e1, e2) ->
      let (array, e1') = expr e1
      and (index, e2') = expr e2 in
      if index <> Int then raise(Failure("index must be integer"))
      else let check_arr_t array = 
        match array with
        ArrayType(t) -> t 
        | _ -> raise(Failure("can only access arrays"))
      in (check_arr_t array, SArrayAccess((array, e1'), (index, e2')))
  | ArrayAppend (e1, e2) -> 
      let (array, e1') = expr e1
      and (typ, element) = expr e2 in
      let check_arr_t array = match array with
        ArrayType(t) -> t
        | _ -> raise(Failure("can only append to arrays"))
      (*check same type here*)
      in (check_arr_t array, SArrayAppend((array, e1'), (typ, element)))
  (*| TrackArrayLiteral(l) -> 
    if List.length l > 0 then 
      let siz = List.length l in 
      let typ = expr(List.nth l 0) in
          let (arraytype, _) = typ in
          let check_array e =
            let (et, e') = expr e in (et, e')
            in let l' = List.map check_array l in
            let types e = 
              let (a, _)  = expr e in 
              if a <> arraytype then 
                      raise(Failure("array of different types")) 
              in let _ = List.map types l in (TrackArrayType(arraytype, List.length l'), STrackArrayLiteral(l')) 


    else raise (Failure ("empty TrackArray initialization not supported"))
  | TrackArrayAccess (e1, e2) as e ->
      let (tatype, e1') = expr e1 in 
      let (index, e2') = expr e2 in
      if index <> Int then raise(Failure("index must be integer"))
      else let check_arr_t tatype = 
        match tatype with
        TrackArrayType(t, _) -> t 
        | _ -> raise(Failure("can only access TrackArray")) in
      let get_size tatype = 
        match tatype with
        TrackArrayType(_, siz) -> siz 
        | _ -> raise(Failure("no size in TrackArray type")) in
      
       (TrackArrayType(check_arr_t tatype, get_size tatype), STrackArrayAccess((tatype, e1'), (index, e2')))
  | TrackArrayInit(i, s) ->
      let expr_id = expr i in
      let (typ, _) = expr_id in 
      let check_type = match typ with
        TrackArrayType(t, _) -> t
        | _ -> raise(Failure("can only initialize a TrackArrayType with new")) in
      let match_type = match check_type with
        Trackpoint -> Trackpoint
        | _ -> raise(Failure("expected Trackpoint, got" ^ string_of_typ check_type)) in 
      (TrackArrayType(match_type, s), STrackArrayInit(expr_id, (Int, SLiteral s))) *)
in

    let check_bool_expr e = 
      let (t', e') = expr e
      and err = "expected Boolean expression in " ^ string_of_expr e
      in if t' != Bool then raise (Failure err) else (t', e') 
    in

    (* Return a semantically-checked statement i.e. containing sexprs *)
    let rec check_stmt = function
        Expr e -> SExpr (expr e)
      | If(p, b1, b2) -> SIf(check_bool_expr p, check_stmt b1, check_stmt b2)
      | For(e1, e2, e3, st) ->
	  SFor(expr e1, check_bool_expr e2, expr e3, check_stmt st) 
      | While(p,s) -> SWhile(check_bool_expr p, check_stmt s)
      | Return e -> let (t, e') = expr e in
        if t = func.typ then SReturn (t, e') 
        else raise (
	  Failure ("return gives " ^ string_of_typ t ^ " expected " ^
		   string_of_typ func.typ ^ " in " ^ string_of_expr e))
	    
	    (* A block is correct if each statement is correct and nothing
	       follows any Return statement.  Nested blocks are flattened. *)
      | Block sl -> 
          let rec check_stmt_list = function
              [Return _ as s] -> [check_stmt s]
            | Return _ :: _   -> raise (Failure "nothing may follow a return")
            | Block sl :: ss  -> check_stmt_list (sl @ ss) (* Flatten blocks *)
            | s :: ss         -> check_stmt s :: check_stmt_list ss
            | []              -> []
          in SBlock(check_stmt_list sl)

    in (* body of check_function *)
    { styp = func.typ;
      sfname = func.fname;
      sformals = func.formals;
      slocals  = func.locals;
      sbody = match check_stmt (Block func.body) with
	SBlock(sl) -> sl
      | _ -> raise (Failure ("internal error: block didn't become a block?"))
    } 
    
  in (globals, List.map check_function functions)
