
/* Ocamlyacc parser for GWiz
   Based on MicroC
   File: gwizparse.mly
   Authors: Ashley Kim, Katherine Duff, Elisa Luo
   PLT Spring 2021
*/

%{
  open Ast
%}

%token SEMI LPAREN RPAREN LBRACE RBRACE LBRAC RBRAC PERIOD COMMA/*seperators*/
%token PLUS MINUS TIMES DIVIDE ASSIGN/*binary operators */
%token NOT EQ NEQ LT LEQ GT GEQ AND OR NEG/*logical operators*/
%token IF ELSE FOR WHILE RETURN INT BOOL DOUBLE STRING VOID TRACKPOINT FILE/*keywords*/
%token <int> LITERAL
%token <bool> BLIT /*boolean literal*/
%token <string> ID SLIT DLIT
%token EOF


%nonassoc NOELSE
%nonassoc ELSE
%left SEMI
%right ASSIGN
%left PERIOD
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE
%right NOT
%right NEG


%start program
%type <Ast.program> program

%%

program:
  decls EOF { $1 }

decls:
    /*nothing*/ {   ([], [])      }
  | decls vdecl { (($2 :: fst $1), snd $1) }
  | decls fdecl { (fst $1, ($2 :: snd $1)) }


fdecl: /* function declaration */
    typ ID LPAREN formals_opt RPAREN LBRACE vdecl_list stmt_list RBRACE
    { { typ = $1;
    fname = $2;
    formals = List.rev $4;
    locals = List.rev $7;
    body = List.rev $8 } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { $1 }

formal_list:
    typ ID                   { [($1,$2)]     }
  | formal_list COMMA typ ID { ($3,$4) :: $1 }

typ:
    INT { Int }
  | DOUBLE { Double }
  | BOOL { Bool }
  | VOID { Void }
  | STRING { String }
  | TRACKPOINT { Trackpoint }
  | FILE { File }
  | typ LBRAC RBRAC { ArrayType($1) }

vdecl_list:
    /*nothing*/ { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl: /*variable declaration */
  typ ID SEMI { ($1, $2) }

stmt_list:
    /*nothing*/ { [] }
  | stmt_list stmt { $2 :: $1 }

stmt: /*statements*/
    expr SEMI { Expr ($1) }
  | RETURN expr_opt SEMI { Return ($2)}
  | LBRACE stmt_list RBRACE { Block(List.rev $2)} /*nesting statements*/
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt { If($3, $5, $7)}
  | FOR LPAREN expr_opt SEMI expr SEMI expr_opt RPAREN stmt { For($3, $5, $7, $9) }
  | WHILE LPAREN expr RPAREN stmt           { While($3, $5)         }

expr_opt: /*when an expression is optional*/
    /*nothing*/ { Noexpr }
  | expr { $1 }

expr:
    LITERAL { Literal($1) }
  | DLIT { DLit($1) }
  | BLIT { BLit($1) }
  | ID { Id($1) }
  | SLIT { SLit($1) }
  | expr PLUS expr { Binop($1, Add, $3) }
  | expr MINUS expr { Binop($1, Sub, $3) }
  | expr TIMES expr { Binop($1, Mult, $3) }
  | expr DIVIDE expr { Binop($1, Div, $3) }
  | expr EQ expr { Binop($1, Equal, $3) }
  | expr NEQ expr { Binop($1, Neq, $3) }
  | expr LT expr { Binop($1, Less, $3) }
  | expr LEQ expr { Binop($1, Leq, $3) }
  | expr GT expr { Binop($1, Greater, $3) }
  | expr GEQ expr { Binop($1, Geq, $3) }
  | expr AND expr { Binop($1, And, $3) }
  | expr OR expr { Binop($1, Or, $3) }
  | MINUS expr %prec NEG{ Unop(Neg, $2) }
  | NOT expr { Unop(Not, $2) }
  | ID ASSIGN expr { Assign($1, $3) }
  | ID LPAREN args_opt RPAREN { Call($1, $3) } /*function calling*/
  | LPAREN expr COMMA expr COMMA expr RPAREN { TrackpointLit($2, $4, $6) }
  | expr PERIOD ID                { TrackpointAccess($1, $3) }
  | LBRAC args_opt RBRAC { ArrayLiteral($2) }
  | expr MINUS LT expr GT { ArrayAccess($1, $4)}
  | expr PLUS LT expr GT { ArrayAppend($1, $4) }



args_opt:
    /*nothing*/ { [] }
  | args_list { List.rev $1 }

args_list:
    expr { [$1] }
  | args_list COMMA expr { $3 :: $1 }
